# -*- coding: utf-8 -*-
'''
A module for running nisms based actions.
'''
# Import python libs
from __future__ import absolute_import
import logging
import os
import os.path
import subprocess
import sys
import tornado.ioloop

# Import salt libs
import salt.payload
import salt.utils
import salt.utils.event
import salt.utils.timed_subprocess
import salt.grains.extra

from nisms_ipc import SMSIPCClient


# Define the module's virtual name
__virtualname__ = 'nisms'

# Define unregister handler name
NISMS_JOB_FUNC_UNREGISTER = 'unregister'
NISMS_JOB_FUNC_REGISTER_HELPER_FILE = 'nisms_register_minion.py'

# Used status codes
NISMS_STATUS_SERVICE_NOT_FOUND = -2147220557
NISMS_STATUS_OPERATION_TIMED_OUT = -2147220448
NISMS_STATUS_INTERNAL_ERROR = -2147418113

# Set up logging
log = logging.getLogger(__name__)


def __virtual__():
    '''
    Overwriting the cmd python module makes debugging modules
    with pdb a bit harder so lets do it this way instead.
    '''
    return __virtualname__


def _fire_master(data, tag):
    '''
    Fire an event to the Salt Master.
    '''
    return salt.utils.event.MinionEvent(__opts__).fire_event(
        {'data': data, 'tag': tag, 'events': None, 'pretag': None},
        'fire_master'
    )


def _mark_job_as_inprogress(jid, id_):
    '''
    Notify the Salt Master that the job is in progress.
    '''
    data = {'jid': jid, 'id': id_}
    tag = 'nisms/job/{0}/inprogress/{1}'.format(jid, id_)
    _fire_master(data, tag)


def _prepare_request_data(current_fun, args, kwargs, minion_id):
    '''
    Prepare the data that will be sent as a job request
    to the NI Minion Agent.
    '''
    fun = __virtualname__ + '.' + current_fun
    return {
        'arg': args,
        'fun': fun,
        'jid': kwargs['__pub_jid'],
        'kwargs': kwargs,
        'id': minion_id
    }


def _query_minionagent(request):
    '''
    Send a job request to the NI Minion Agent
    and receive a response with the job return info.
    '''
    io_loop = tornado.ioloop.IOLoop()
    client = None
    try:
        uri = 7890
        client = SMSIPCClient(uri, io_loop=io_loop)
        try:
            io_loop.run_sync(
                lambda: client.connect(timeout=5))
        except Exception:  # pylint: disable=broad-except
            log.error('Connection failed to: {0}'.format(uri))
            return {'statusCode': NISMS_STATUS_SERVICE_NOT_FOUND}

        try:
            log.info('About to send request to: {0}'.format(uri))
            response = io_loop.run_sync(
                lambda: client.query(request))
        except Exception as exc:  # pylint: disable=broad-except
            log.error('Request to: {0} failed. Exception = {1}'.format(uri, exc))
            return {'statusCode': NISMS_STATUS_OPERATION_TIMED_OUT}
    finally:
        if client is not None:
            client.close()
        io_loop.close()

    log.info('Response received!')
    return response


def _unregister_minion():
    '''
    Invoke the register script's 'unregister'
    capabilities to unregister the minion.
    '''
    helper_file_path = os.path.join(
        os.path.dirname(
            os.path.abspath(__file__)
        ),
        NISMS_JOB_FUNC_REGISTER_HELPER_FILE
    )
    # Start non-blocking python process. This call cannot block
    # because the spawned process can restart the minion
    result = subprocess.Popen(['python', helper_file_path, '--unregister'])
    # 'result.returncode' is most likely 'None' meaning in progress unless it
    # couldn't execute the script.
    return result.returncode


def _restart_minion():
    '''
    Invoke the register script's 'reset'
    capabilities to cleanly restart the minion daemon.
    '''

    helper_file_path = os.path.join(
        os.path.dirname(
            os.path.abspath(__file__)),
        NISMS_JOB_FUNC_REGISTER_HELPER_FILE
    )
    # Start non-blocking python process. This call cannot block
    # because the spawned process can restart the minion
    result = subprocess.Popen(
        [
            'python',
            helper_file_path,
            '--reset',
            '--delay',
            '5'
        ]
    )
    # 'result.returncode' is most likely 'None' meaning in progress unless it
    # couldn't execute the script.
    return result.returncode


def _generic_func_impl(current_fun, args, kwargs):
    '''
    This is the generic function implementation
    that calls into the NI Minion Agent.

    Most top-level public APIs would call into this function.
    '''
    jid = kwargs['__pub_jid']
    minion_id = __grains__['id']
    log.info(
        '_generic_func_impl: minion_id = {0}, '
        'jid = {1}, fun = {2}'.format(
            minion_id, jid, kwargs['__pub_fun']
        )
    )

    if current_fun == NISMS_JOB_FUNC_UNREGISTER:
        response = _unregister_minion()
    else:
        response = _query_minionagent(
            _prepare_request_data(current_fun, args, kwargs, minion_id)
            )
    return response


def delete_resource(*args, **kwargs):
    '''
    Delete a given resource
    '''
    current_fun = sys._getframe().f_code.co_name  # pylint: disable=protected-access
    return _generic_func_impl(current_fun, args, kwargs)


def get_info(*args, **kwargs):
    '''
    Get system information
    '''
    current_fun = sys._getframe().f_code.co_name  # pylint: disable=protected-access
    return _generic_func_impl(current_fun, args, kwargs)


def get_props(*args, **kwargs):
    '''
    Get System API properties
    '''
    current_fun = sys._getframe().f_code.co_name  # pylint: disable=protected-access
    return _generic_func_impl(current_fun, args, kwargs)


def populate_props(*args, **kwargs):
    '''
    Get specific System API properties
    '''
    current_fun = sys._getframe().f_code.co_name  # pylint: disable=protected-access
    return _generic_func_impl(current_fun, args, kwargs)


def set_props(*args, **kwargs):
    '''
    Set System API properties
    '''
    current_fun = sys._getframe().f_code.co_name  # pylint: disable=protected-access
    response = _generic_func_impl(current_fun, args, kwargs)

    should_restart_minion = False
    if 'networkSettingsChange' in response:
        should_restart_minion = response['networkSettingsChange']
        # Do not save 'networkSettingsChange' in the master.
        del response['networkSettingsChange']

    if should_restart_minion and not salt.utils.is_windows():
        # The script that restarts the minion does not currently work on
        # Windows. SysApi doesn't typically support changing network settings
        # on Windows, but it is possible that an external expert might claim
        # that a property bag supports 'mxSysServiceType_LocalNetInterface' so
        # we keep the Windows check in place.
        result = _restart_minion()
        if result is not None and result != 0:
            if response['statusCode'] >= 0:
                response['statusCode'] = NISMS_STATUS_INTERNAL_ERROR

    return response


def reset_device(*args, **kwargs):
    '''
    Reset a device
    '''
    current_fun = sys._getframe().f_code.co_name  # pylint: disable=protected-access
    return _generic_func_impl(current_fun, args, kwargs)


def restart(*args, **kwargs):
    '''
    Restart the minion
    '''
    current_fun = sys._getframe().f_code.co_name  # pylint: disable=protected-access
    return _generic_func_impl(current_fun, args, kwargs)


def self_cal(*args, **kwargs):
    '''
    Calibrate a device
    '''
    current_fun = sys._getframe().f_code.co_name  # pylint: disable=protected-access
    return _generic_func_impl(current_fun, args, kwargs)


def self_test(*args, **kwargs):
    '''
    Self-test a device
    '''
    current_fun = sys._getframe().f_code.co_name  # pylint: disable=protected-access
    return _generic_func_impl(current_fun, args, kwargs)


def unregister(*args, **kwargs):
    '''
    Unregister the minion
    '''
    current_fun = sys._getframe().f_code.co_name  # pylint: disable=protected-access
    return _generic_func_impl(current_fun, args, kwargs)
