#!/usr/bin/python
# -*- coding: utf-8 -*-

'''
This script is used to register a minion to a master
or unregister a minion from a master
'''
from __future__ import absolute_import
import argparse  # pylint: disable=minimum-python-version
import os
import os.path
import subprocess
import sys
import time
import yaml


def is_windows():
    '''
    Return True if the system platform is Windows. Return False otherwise.
    '''
    return sys.platform.startswith('win')


def is_linux():
    '''
    Return True if the system platform is Linux. Return False otherwise.
    '''
    return sys.platform.startswith('linux')


if is_windows():
    import win32con  # pylint: disable=wrong-import-position
    import win32file  # pylint: disable=wrong-import-position
    import pywintypes  # pylint: disable=wrong-import-position
    from win32com.shell import shellcon, shell  # pylint: disable=no-name-in-module, import-error, wrong-import-position
else:
    # Non-Windows platforms are typically POSIX compliant.
    import fcntl  # pylint: disable=import-error, wrong-import-position


class FileLockException(Exception):
    '''
    Class for file locking exceptions
    '''
    # Error codes:
    LOCK_FAILED = 1


class FileLock(object):
    '''
    Class that implements file-based locking
    '''
    def __init__(self, filename):
        self.filename = filename
        # This will create it if it does not exist already
        self.handle = open(filename, 'w')
        if is_windows():
            self.hfile = win32file._get_osfhandle(self.handle.fileno())  # pylint: disable=no-member, protected-access
            self.overlapped = pywintypes.OVERLAPPED()  # pylint: disable=no-member

    def acquire(self):
        '''
        Acquire the file-based lock.
        Will raise a FileLockException if the lock cannot be acquired.
        '''
        if is_windows():
            try:
                win32file.LockFileEx(  # pylint: disable=no-member
                    self.hfile,
                    win32con.LOCKFILE_EXCLUSIVE_LOCK,
                    0,
                    -0x10000,
                    self.overlapped
                )
            except pywintypes.error as exc_value:  # pylint: disable=no-member
                # error: (33, 'LockFileEx',
                #         'The process cannot access the file because another
                #          process has locked a portion of the file.')
                if exc_value[0] == 33:
                    raise FileLockException(
                        FileLockException.LOCK_FAILED,
                        exc_value[2]
                    )
                else:
                    raise
        else:
            try:
                fcntl.flock(self.handle, fcntl.LOCK_EX)
            except IOError as exc_value:
                #  IOError: [Errno 11] Resource temporarily unavailable
                if exc_value[0] == 11:
                    raise FileLockException(
                        FileLockException.LOCK_FAILED,
                        exc_value[1]
                    )
                else:
                    raise

    def release(self):
        '''
        Release the file-based lock.
        The lock must already be previously acquired via acquire().
        '''
        if is_windows():
            try:
                win32file.UnlockFileEx(  # pylint: disable=no-member
                    self.hfile, 0, -0x10000, self.overlapped
                )
            except pywintypes.error as exc_value:  # pylint: disable=no-member
                if exc_value[0] == 158:
                    # error: (158, 'UnlockFileEx',
                    #         'The segment is already unlocked.')
                    # To match the 'posix' implementation, silently ignore
                    # this error
                    pass
                else:
                    raise
        else:
            fcntl.flock(self.handle, fcntl.LOCK_UN)

    def __del__(self):
        self.handle.close()

    def __enter__(self):
        self.acquire()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.release()


class MasterListArgumentAction(argparse.Action):  # pylint: disable=too-few-public-methods
    '''
    Class that defines an action to take for the Master List
    from within 'argparse.ArgumentParser'.
    '''
    def __call__(self, parser, namespace, values, option_string=None):
        if isinstance(values, str):
            values = values.split(',')
        if 'master_list' in namespace:
            namespace.master_list.extend(values)
        else:
            setattr(namespace, 'master_list', values)


def get_lock_file_path():
    '''
    Return path for concurrency control lock file.
    eg: '/var/lock/nisms_register_minion.py.lock'
    '''
    if is_windows():
        # Use the Windows services temp directory which is typically:
        # 'C:\Windows\Temp'
        lock_file = shell.SHGetFolderPath(0, shellcon.CSIDL_WINDOWS, 0, 0)
        lock_file = os.path.join(lock_file, 'Temp')
    elif is_linux():
        lock_file = '/var/lock'
    else:
        raise RuntimeError('Unsupported platform')
    lock_filename = '{0}.lock'.format(os.path.basename(__file__))
    lock_file = os.path.join(lock_file, lock_filename)
    return lock_file


def get_salt_config_dir():
    '''
    Return Salt configuration directory.
    eg: '/etc/salt'
    '''
    if is_windows():
        config_dir = shell.SHGetFolderPath(
            0, shellcon.CSIDL_COMMON_APPDATA, 0, 0
        )
        config_dir = os.path.join(
            config_dir, 'National Instruments', 'salt', 'conf'
        )
        return config_dir
    elif is_linux():
        return '/etc/salt'
    else:
        raise RuntimeError('Unsupported platform')


def get_salt_minion_config_subdir():
    '''
    Return Salt Minion configuration subdirectory.
    eg: '/etc/salt/minion.d'
    '''
    config_dir = get_salt_config_dir()
    config_dir = os.path.join(config_dir, 'minion.d')
    return config_dir


def get_salt_minion_master_config_file():  # pylint: disable=invalid-name
    '''
    Return Salt Minion's configuration file that specifies which master to use.
    eg: '/etc/salt/minion.d/master.conf'
    '''
    config_file = get_salt_minion_config_subdir()
    config_file = os.path.join(config_file, 'master.conf')
    return config_file


def get_salt_pki_dir():
    '''
    Return Salt Public Key Infrastructure (PKI) directory.
    eg: '/etc/salt/pki'
    '''
    pki_dir = get_salt_config_dir()
    pki_dir = os.path.join(pki_dir, 'pki')
    return pki_dir


def get_salt_pki_minion_subdir():
    '''
    Return Salt PKI Minion-specific subdirectory.
    eg: '/etc/salt/pki/minion'
    '''
    pki_dir = get_salt_pki_dir()
    pki_dir = os.path.join(pki_dir, 'minion')
    return pki_dir


def get_salt_pki_minion_master_pubkey_file():  # pylint: disable=invalid-name
    '''
    Return Salt PKI public key file that is generated by the salt-master.
    eg: '/etc/salt/pki/minion/minion_master.pub'
    '''
    pki_file = get_salt_pki_minion_subdir()
    pki_file = os.path.join(pki_file, 'minion_master.pub')
    return pki_file


def uniquify_list(list_arg):
    '''
    Return a list of the unique elements in the list argument while
    preserving their order.
    '''
    added = set()
    ordered_list = []
    for element in list_arg:
        if element in added:
            continue
        added.add(element)
        ordered_list.append(element)
    return ordered_list


def set_salt_minion_master_config(args):
    '''
    Changes the Master hostname and fingerprint to what is specified by args.
    Return old_master, master_changed
    old_master: The Master hostname or list of addresses used previously.
                Empty string if none was found.
    master_changed: True if the new Master hostname is different from the
                    previous one. False otherwise.
    '''
    if args.unregister is True:
        # Set the Master address to an unroutable address (empty string) and
        # set the master fingerprint to an empty string
        new_master = ''
        master_finger = ''
    else:
        # The master_list argument is always a list.
        # A list is required for the yaml dump and to preserve order.
        new_master = uniquify_list(args.master_list)  # pylint: disable=redefined-variable-type
        if len(new_master) == 1:
            new_master = new_master[-1]
        master_finger = args.master_finger

    master_changed = True
    old_master = ''

    config_file = get_salt_minion_master_config_file()
    if os.path.isfile(config_file):
        with open(config_file, 'r') as fp:  # pylint: disable=invalid-name
            config_data = yaml.load(fp)

        if isinstance(config_data, dict):
            if 'master' in config_data:
                old_master = config_data['master']
                if new_master == old_master:
                    master_changed = False
        else:
            config_data = {}
    else:
        config_data = {}

    config_data['master_finger'] = master_finger
    if master_changed:
        config_data['master'] = new_master
        if isinstance(new_master, str):
            config_data['master_type'] = 'str'
        else:
            config_data['master_type'] = 'failover'
            # Avoid this error in the log:
            #     [CRITICAL][2205] 'master_type' set to 'failover' but
            #     'retry_dns' is not 0. Setting 'retry_dns' to 0 to
            #     failover to the next master on DNS errors.
            config_data['retry_dns'] = 0
    if args.unregister is True:
        config_data['master_type'] = 'disable'
    with open(config_file, 'w') as fp:  # pylint: disable=invalid-name
        yaml.dump(config_data, fp, default_flow_style=False)

    return old_master, master_changed


def clear_pki_minion_master_cache():
    '''
    Removes the Minion's PKI cache of the Master public key.
    Meant to be called if the Minion ID or Master hostname has changed.
    '''
    pki_file = get_salt_pki_minion_master_pubkey_file()
    if os.path.isfile(pki_file):
        os.remove(pki_file)


def stop_minion_if_running():
    '''
    This will stop the Salt Minion if it is running.
    This will do nothing if the Salt Minion is not running.
    '''
    if is_windows():
        retcode = subprocess.call(['net', 'stop', 'nisaltminion'])
        # retcode of 2 means that the service is not running
        if retcode != 0 and retcode != 2:
            sys.stderr.write(
                '{0}: Could not stop the salt-minion service. '
                'Return code: {1}.'.format(
                    os.path.basename(__file__),
                    retcode
                )
            )
            sys.exit(1)
    else:
        # This will still return 0 even if the daemon is already stopped
        retcode = subprocess.call(['/etc/init.d/salt-minion', 'stop'])
        if retcode != 0:
            sys.stderr.write(
                '{0}: Could not stop the salt-minion daemon. '
                'Return code: {1}.'.format(
                    os.path.basename(__file__),
                    retcode
                )
            )
            sys.exit(1)


def start_minion_if_not_running():
    '''
    This will start the Salt Minion if it is not running.
    This will do nothing if the Salt Minion is running.
    '''
    if is_windows():
        retcode = subprocess.call(['net', 'start', 'nisaltminion'])
        # retcode of 2 means that the service is already running
        if retcode != 0 and retcode != 2:
            sys.stderr.write(
                '{0}: Could not start the salt-minion service. '
                'Return code: {1}.'.format(
                    os.path.basename(__file__),
                    retcode
                )
            )
            sys.exit(1)
    else:
        # This will still return 0 even if the daemon is already running
        retcode = subprocess.call(['/etc/init.d/salt-minion', 'start'])
        if retcode != 0:
            sys.stderr.write(
                '{0}: Could not restart the salt-minion daemon. '
                'Return code: {1}.'.format(
                    os.path.basename(__file__),
                    retcode
                )
            )
            sys.exit(1)


def parse_args():
    '''
    This will parse the command-line arguments passed in.
    Returns a dictionary of the parsed command-line arguments.

    The master and master-list arguments are stored in the same list to
    preserve order.
    '''
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--master',
        action='append',
        dest='master_list',
        default=[],
        help='An address of the salt-master to connect to. Must also'
             'specify a master finger.'
    )
    parser.add_argument(
        '--master-list',
        action=MasterListArgumentAction,
        dest='master_list',
        default=[],
        help='A comma separated list of addresses of the salt-master to '
             'connect to. Must also specify a master finger.'
    )
    parser.add_argument(
        '--unregister',
        action='store_true',
        dest='unregister',
        default=False,
        help='Unregister the current salt-master'
    )
    parser.add_argument(
        '--reset',
        action='store_true',
        dest='reset',
        default=False,
        help='Reset the connection to the existing salt-master'
    )
    parser.add_argument(
        '--delay',
        dest='delay',
        default=0,
        type=int,
        help='Number of seconds to delay before starting'
    )
    parser.add_argument(
        '--master-finger',
        dest='master_finger',
        default='',
        type=str,
        help='Fingerprint of the master public key used to validate the '
             'identity of the Salt master before the initial key '
             'exchange. Must also specify a master or master list.'
    )
    args = parser.parse_args()

    new_master = uniquify_list(args.master_list)
    new_master_len = len(new_master)

    if args.reset:
        if (args.unregister is True or new_master_len != 0 or
                args.master_finger):
            parser.print_help()
            sys.exit(1)
    else:
        if (args.unregister is True and
                (new_master_len != 0 or args.master_finger)):
            parser.print_help()
            sys.exit(1)

        if (args.unregister is False and
                not args.master_finger and
                new_master_len == 1 and
                '' in new_master):
            args.unregister = True
            args.master_list = []
        elif (args.unregister is False and
              (not args.master_finger or new_master_len == 0)):
            parser.print_help()
            sys.exit(1)

    return args


def main():
    '''
    This is the main entry point for this script.
    '''
    if not is_windows() and not is_linux():
        sys.stderr.write(
            '{0}: Unsupported platform. Only Windows and Linux are '
            'supported.'.format(
                os.path.basename(__file__)
            )
        )
        sys.exit(1)

    args = parse_args()

    if args.delay:
        time.sleep(args.delay)

    lock_file = get_lock_file_path()
    lock = FileLock(lock_file)
    with lock:
        if not args.reset:
            set_salt_minion_master_config(args)

        stop_minion_if_running()
        if not args.reset:
            clear_pki_minion_master_cache()
        start_minion_if_not_running()

        sys.exit(0)


if __name__ == '__main__':
    main()
